/************************************************************************/
/* FSD.C, Floyd-Steinberg Dither example written by Chem of Hubris 1998 */
/* release 2. Compiles on DJGPP v2 and Watcom C 10+. PUBLIC DOMAIN, no  */
/* responsibilities, no support, no nothing. Use it at your own risk.   */
/************************************************************************/
/* compile :                                                            */
/*   gcc -O3 -Wall -g -o fsd.exe fsd.c                                  */
/*   wcl386 /ox fsd.c                                                   */
/************************************************************************/


/*
  The Idea of FS-Dithering

  You have a perfect palette, for example a 24-bit image, and then you
  have a non-perfect palette, for instance a quantized 8-bit or even a 2
  color black & white palette. The Floyd-Steinberg dithering algorithm can
  greatly improve the quality of the quantized image and thus it is a
thing
  worth learning.

  It's simple. For each pixel let i be the color value from the original
  image plus the error for the pixel. Find the closest color to i in your
  palette and output it to the destination. The new error is the signed
  difference between the closest color and i.

  The error is then distributed for the pixels near the currently
  processed pixel. In the classical FS-dither 7/16 of the error goes to
  the horizontally next pixel (==right), 5/16 to the vertically next pixel
  (==down), 1/16 to the pixel next to the vertically next pixel (==down &
  right) and 3/16 to the pixel before the next to the vertically next
pixel
  (==down & left). Previously I claimed the original distribution to be
  3/8-3/8-1/8-1/8 which it isn't(?).

       +----+----+
       | 0  | 1  |  0 = the pixel being processed
       |    |    |  1 = the horizontally next pixel (7/16)
  +----+----+----+  2 = the vertically next pixel (5/16)
  | 4  | 2  | 3  |  3 = the pixel next to the vertically next pixel (1/16)
  |    |    |    |  4 = the pixel before the vertically next pixel (3/16)
  +----+----+----+

  It's not absolutely necessary to distribute all of the error and you
  don't have to use just those "weights". If you are looking for speed,
  you may want to use a simple 1-1-1-1 or even 1-1-1 -style dithering.

  What's new since release 1:
    - New weights as described above.
    - Memcpy replaced with a pointer swap.
    - Right to left-processing for odd lines.
    - Raw 320x200 24bit rgbrgb image loading support. You can convert
      a image to this format for example with psp.
    - Local k-stylish quantization added [slooow..]
*/


/*
  If use_picture is #defined, a 320x200 24bit rgbrgb image will
  be loaded from PIC.RAW.
*/
/* #define use_picture */

/*
  If quantize is #defined, the program will quantize the example image
  and add 2 steps to the slideshow. Quantization is VERY SLOW though.
*/
/* #define quantize */
#define quantize_passes 4

#define XSIZE 320L
#define YSIZE 200L

#define uchar unsigned char
#define uint unsigned int
#define sint signed int
#define ushort unsigned short
#define sshort signed short

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <malloc.h>
#include <conio.h>

#ifdef DJGPP
#include <go32.h>
#include <sys\farptr.h>
#include <pc.h>
#endif

#ifdef __WATCOMC__
/*
  note: this _farpokeb for wc will only work if the selector is _dos_ds
*/
#define _farpokeb(a, b, c) *((uchar*)(b))=((uchar)c);
#define _dos_ds 0
#define inline
void textmode(int _mode);
#pragma aux textmode = "int 010h" parm [eax];
#endif

#define setpixel(x, y, c) \
  _farpokeb(_dos_ds, 0xA0000 + (x) + (y) * XSIZE, c)

typedef struct
{
  uchar r, g, b;
} rgb;

typedef struct
{
  uint r, g, b;
} brgb;

#ifdef quantize
uint square[512];
uint show_quantize;
rgb *show_palette;

inline uint find_closest_color(uchar r, uchar g, uchar b, rgb *palette)
{
  register uint i, bd = 0x7FFFFFFF, bdp = 257, nd;

  for (i = 0L; i < 256L; i++) {
    nd = (square[256L + (sint)palette[i].r - (sint)r] +
          square[256L + (sint)palette[i].g - (sint)g] +
          square[256L + (sint)palette[i].b - (sint)b]);
    if(nd <= bd) {
      bd = nd;
      bdp = i;
    }
  }

  return bdp;
}

void lkm_quantize(rgb *virtscr, rgb *palette)
{
  uint i, pass, closest;
  brgb *pb_new_pos = (brgb*)malloc(sizeof(brgb) * 256L);
  uint *pb_new_pos_cnt = (uint*)malloc(sizeof(uint) * 256L);

  for (pass = 0; pass < quantize_passes; pass++) {
    memset(pb_new_pos, 0, sizeof(brgb) * 256L);
    memset(pb_new_pos_cnt, 0, sizeof(uint) * 256L);

    for (i = 0; i < XSIZE * YSIZE; i++) {
      closest = find_closest_color(virtscr[i].r,
                                   virtscr[i].b,
                                   virtscr[i].g,
                                   palette);
      pb_new_pos[closest].r += virtscr[i].r;
      pb_new_pos[closest].g += virtscr[i].g;
      pb_new_pos[closest].b += virtscr[i].b;
      pb_new_pos_cnt[closest]++;
    }

    for (i = 0; i < 256; i++) {
      if(pb_new_pos_cnt[i] == 0) {
        closest = rand() % (XSIZE * YSIZE);
        palette[i].r = virtscr[closest].r;
        palette[i].g = virtscr[closest].g;
        palette[i].b = virtscr[closest].b;
      } else {
        palette[i].r = pb_new_pos[i].r / pb_new_pos_cnt[i];
        palette[i].g = pb_new_pos[i].g / pb_new_pos_cnt[i];
        palette[i].b = pb_new_pos[i].b / pb_new_pos_cnt[i];
      }
    }

  }
}
#endif

inline uchar get_color_index(uchar r, uchar g, uchar b)
{
#ifdef quantize
  if (!show_quantize)
#endif
    return ((((uint)r * 7L) / 255L)) +
           ((((uint)g * 7L) / 255L) << 3) +
           ((((uint)b * 3L) / 255L) << 6);
#ifdef quantize
  return find_closest_color(r, g, b, show_palette);
#endif
}

inline uchar klip(sint o)
{
  if(o < 0)   return 0;
  if(o > 255) return 255;
  return (uchar)o;
}

void floyd_steinberg_flip(rgb *virtscr, rgb *palette)
{
  sint nr, ng, nb, er, eg, eb, r, g, b;
  sint *uvr = (sint*)malloc((XSIZE + 10) * sizeof(sint));
  sint *uvg = (sint*)malloc((XSIZE + 10) * sizeof(sint));
  sint *uvb = (sint*)malloc((XSIZE + 10) * sizeof(sint));
  sint *nvr = (sint*)malloc((XSIZE + 10) * sizeof(sint));
  sint *nvg = (sint*)malloc((XSIZE + 10) * sizeof(sint));
  sint *nvb = (sint*)malloc((XSIZE + 10) * sizeof(sint));
  sint *ooo;
  uint a, nyy, pw, nw;
  sint xb, d, c;
  rgb co;

  memset(nvr, 0, sizeof(sint) * (XSIZE + 10));
  memset(nvg, 0, sizeof(sint) * (XSIZE + 10));
  memset(nvb, 0, sizeof(sint) * (XSIZE + 10));

  nr = 0;
  ng = 0;
  nb = 0;

  for (a = 0; a < 200; a++) {
    memset(uvr, 0, sizeof(sint) * (XSIZE + 10));
    memset(uvg, 0, sizeof(sint) * (XSIZE + 10));
    memset(uvb, 0, sizeof(sint) * (XSIZE + 10));

    /* if the line number is odd (1,3,5,7..), then do the dither from */
    /* right to left instead of left to right. */
    if ((a & 1) == 0) {
      pw = 3;
      nw = 1;
      d = 1;
      xb = 0;
    } else {
      pw = 1;
      nw = 3;
      d = -1;
      xb = XSIZE - 1;
    }

    for (c = xb; (c < XSIZE) && (c >= 0); c += d) {
      co = virtscr[c + a * XSIZE];

      /* the color value (rgb) is the color value for this pixel */
      /* + the error that this pixel has gotten in the distribution */
      r = ((sint)co.r * 16L + nr + nvr[c + 1]) / 16L;
      g = ((sint)co.g * 16L + ng + nvg[c + 1]) / 16L;
      b = ((sint)co.b * 16L + nb + nvb[c + 1]) / 16L;

      nyy = get_color_index(klip(r), klip(g), klip(b));

      setpixel(c, a, nyy);

      er = r - (sint)palette[nyy].r;
      eg = g - (sint)palette[nyy].g;
      eb = b - (sint)palette[nyy].b;

      nr = er * 7L;
      uvr[c] += er * pw;
      uvr[c + 1] += er * 5L;
      uvr[c + 2] += er * nw;

      ng = eg * 7L;
      uvg[c] += eg * pw;
      uvg[c + 1] += eg * 5L;
      uvg[c + 2] += eg * nw;

      nb = eb * 7L;
      uvb[c] += eb * pw;
      uvb[c + 1] += eb * 5L;
      uvb[c + 2] += eb * nw;
    }
    ooo = nvr; nvr = uvr; uvr = ooo;
    ooo = nvg; nvg = uvg; uvg = ooo;
    ooo = nvb; nvb = uvb; uvb = ooo;
  }
  free(uvr);
  free(uvg);
  free(uvb);
  free(nvr);
  free(nvg);
  free(nvb);
}

int main(void)
{
  sint i;
  rgb *virtscr = (rgb*)malloc(sizeof(rgb) * XSIZE * YSIZE);
  rgb *palette = (rgb*)malloc(sizeof(rgb) * 256L);
#ifdef quantize
  rgb *qpalette = (rgb*)malloc(sizeof(rgb) * 256L);
#endif
#ifdef use_picture
  FILE *infile;
#endif

  printf("\nthis program will show you different versions of a picture.
after\n");
  printf("each version you must press any key. a desciption of pictures:
\n");
  printf(" 1) picture uses a standard palette. (3 bits for r&g, 2 for
b)\n");
  printf(" 2) same as 1 but with floyd-steinberg dither.\n");
#ifdef quantize
  printf(" 3) picture uses a quantized (local k stylish) palette.\n");
  printf(" 4) same as 3 but with floyd-steinberg dither.\n");
#endif
  printf("\nplease wait, working..");
  fflush(stdout);

#ifdef quantize
  for (i = -256; i < 256; i++) square[256 + i] = (i * i);
  show_quantize = 0;
  show_palette  = palette;
#endif

  memset(virtscr, 0, sizeof(rgb) * XSIZE * YSIZE);
#ifdef use_picture
  infile = fopen("pic.raw", "rb");
  rewind(infile);
  fread(virtscr, 1, XSIZE * YSIZE * 3L, infile);
  fclose(infile);
#else
  /* generate a test image */
  for (i = 0; i < XSIZE * YSIZE; i++) {
    virtscr[i].r = klip(280.0 - i * 255L / (XSIZE * YSIZE - 1L));
    virtscr[i].g = klip(280.0 - (i % XSIZE) * 255L / (XSIZE - 1));
    virtscr[i].b = klip(
      floor(280.0 - sqrt(pow((i % XSIZE) - (XSIZE * 47L / 50L), 2) +
        pow((i / XSIZE) - (YSIZE * 19L / 20L), 2)) * 255.0 / 180.0
      )
    );
  }
#endif

  for (i = 0; i < 256; i++) {
    palette[i].r = ((i & 7) * 255L) / 7;
    palette[i].g = (((i >> 3) & 7) * 255L) / 7;
    palette[i].b = (((i >> 6) & 3) * 255L) / 3;
  }

#ifdef quantize
  memcpy(qpalette, palette, sizeof(rgb) * 256);
  lkm_quantize(virtscr, qpalette);
#endif

  printf("ok! press a key. \n");
  fflush(stdout);
  getch();

/*
  The higher the resolution, the better the dithered image will look.
  I tested this by writing a quick SVGA-support, and "oh boy" was it
  beautiful. The SVGA-version is not included though, as it got pretty
  messy and was DJGPP-only.

  But even this MCGA example will demonstrate the idea.
*/
  textmode(0x13);

  for (i = 0; i < 256; i++) {
    outp(0x3C8, (uchar)i);
    outp(0x3C9, (uchar)((uint)palette[i].r * 63L / 255L));
    outp(0x3C9, (uchar)((uint)palette[i].g * 63L / 255L));
    outp(0x3C9, (uchar)((uint)palette[i].b * 63L / 255L));
  }

  for (i = 0; i < XSIZE * YSIZE; i++) setpixel(i, 0,
    get_color_index(virtscr[i].r, virtscr[i].g, virtscr[i].b));

  getch();

  floyd_steinberg_flip(virtscr, palette);

  getch();

#ifdef quantize
  show_quantize = 1;
  show_palette = qpalette;

  for (i = 0; i < 256; i++) {
    outp(0x3C8, (uchar)i);
    outp(0x3C9, (uchar)((uint)qpalette[i].r * 63L / 255L));
    outp(0x3C9, (uchar)((uint)qpalette[i].g * 63L / 255L));
    outp(0x3C9, (uchar)((uint)qpalette[i].b * 63L / 255L));
  }

  for (i = 0; i < XSIZE * YSIZE; i++) setpixel(i, 0,
    get_color_index(virtscr[i].r, virtscr[i].g, virtscr[i].b));

  getch();

  floyd_steinberg_flip(virtscr, qpalette);

  getch();
#endif

  textmode(0x3);

  free(virtscr);
  free(palette);

#ifdef quantize
  free(qpalette);
#endif

  return 0;
  /* mailto:kbjorklu@cc.hut.fi */
}
